import unittest

import pandas as pd
import pytest

import riptable as rt

# N.B. TL;DR We have to import the actual implementation module to override the module global
#      variable "tm.N" and "tm.K".
#      In pandas 1.0 they move the code from pandas/util/testing.py to pandas/_testing.py.
#      The "import pandas.util.testing" still works but because it doesn't contain the actual code
#      our attempt to override the "tm.N" and "tm.K" will not change the actual value for
#      makeTimeDataFrame, which will produce data with different shape and make the test
#      "test_accum_table" fail. Maybe we want to reconsider using the pandas internal testing utils.
try:
    import pandas._testing as tm
except ImportError:
    import pandas.util.testing as tm

from numpy.testing import (
    assert_almost_equal,
    assert_array_almost_equal,
    assert_array_equal,
)

from riptable import *
from riptable.rt_datetime import DateTimeNano
from riptable.rt_numpy import arange, QUANTILE_METHOD_NP_KW

# To create AccumTable test data
from riptable.Utils.pandas_utils import dataset_from_pandas_df

tm.N = 3
tm.K = 5


class Accum2_Test(unittest.TestCase):
    """
    TODO: add more tests for different types
    """

    def test_accum2(self):
        c = cut(arange(10), 3)
        self.assertTrue(sum(c._np - FA([1, 1, 1, 1, 2, 2, 2, 3, 3, 3])) == 0)

        c = cut(arange(10.0), 3)
        self.assertTrue(sum(c._np - FA([1, 1, 1, 1, 2, 2, 2, 3, 3, 3])) == 0)

        c = cut(arange(11), 3)
        self.assertTrue(sum(c._np - FA([1, 1, 1, 1, 2, 2, 2, 3, 3, 3, 3])) == 0)

        c = cut(FA([2, 4, 6, 8, 10]), FA([0, 2, 4, 6, 8, 10]))
        self.assertTrue(sum(c._np - FA([1, 2, 3, 4, 5])) == 0)

        c = cut(
            FA([2, 4, 6, 8, 10]),
            FA([0, 2, 4, 6, 8, 10]),
            labels=["a", "b", "c", "d", "e"],
        )
        self.assertTrue(sum(c._np - FA([1, 2, 3, 4, 5])) == 0)

    def test_qcut(self):
        c = qcut(arange(10), 3)
        self.assertTrue(sum(c._np - FA([2, 2, 2, 2, 3, 3, 3, 4, 4, 4])) == 0)

        c = qcut(arange(11), 3)
        self.assertTrue(sum(c._np - FA([2, 2, 2, 2, 3, 3, 3, 4, 4, 4, 4])) == 0)

        c = qcut(range(5), 3, labels=["good", "medium", "bad"])
        self.assertTrue(sum(c._np - FA([2, 2, 3, 4, 4])) == 0)

        c = cut(
            FA([2, 4, 6, 8, 10]),
            FA([0, 2, 4, 6, 8, 10]),
            labels=["a", "b", "c", "d", "e"],
        )

    def test_cut_errors(self):
        with self.assertRaises(ValueError):
            c = cut(
                FA([2, 4, 6, 8, 10]),
                FA([0, 2, 4, 6, 8, 10]),
                labels=["a", "b", "c", "d", "e", "f"],
            )

    def test_simple_cats(self):
        data = arange(1, 6) * 10
        colnames = FastArray(["a", "b", "c", "d", "e"])
        c1 = Categorical(colnames)
        c2 = Categorical(arange(5))

        # no filter
        ac = Accum2(c2, c1)
        result = ac.sum(data)
        self.assertEqual(result._ncols, 7)
        for i, colname in enumerate(colnames):
            arr = result[colname]
            self.assertEqual(arr[i], data[i])

    def test_simple_cats_filter_accum(self):
        data = arange(1, 6) * 10
        colnames = FastArray(["a", "b", "c", "d", "e"])
        c1 = Categorical(colnames)
        c2 = Categorical(arange(5))

        # filtered accum object
        ac = Accum2(c2, c1, showfilter=True)
        result = ac.sum(data)
        self.assertEqual(result._ncols, 8)
        for i, colname in enumerate(colnames):
            arr = result[colname]
            self.assertEqual(arr[i + 1], data[i])

    def test_simple_cats_filter_operation(self):
        data = arange(1, 6) * 10
        colnames = FastArray(["a", "b", "c", "d", "e"])
        c1 = Categorical(colnames)
        c2 = Categorical(arange(5))

        # filtered operation
        ac = Accum2(c2, c1)
        result = ac.sum(data, showfilter=True)
        self.assertEqual(result._ncols, 8)
        for i, colname in enumerate(colnames):
            arr = result[colname]
            self.assertEqual(arr[i + 1], data[i])

    def test_multikey_cats(self):
        unsorted_str = FastArray(["c", "e", "b", "d", "a"])
        ints = arange(1, 6) * 10
        data = np.random.rand(5) * 10

        # unsorted no filter
        c1 = Categorical([unsorted_str, ints])
        c2 = Categorical([unsorted_str, ints])
        ac = Accum2(c2, c1)
        result = ac.sum(data)
        self.assertEqual(result._ncols, 8)
        for i, key1 in enumerate(unsorted_str):
            k1 = bytes.decode(key1)
            k2 = ints[i]
            full_colname = "('" + k1 + "', " + str(k2) + ")"
            arr = result[full_colname]
            self.assertEqual(arr[i], data[i])

        # sorted no filter
        sortidx = np.argsort(unsorted_str)
        sorted_str = unsorted_str[sortidx]
        sorted_ints = ints[sortidx]
        sorted_data = data[sortidx]
        c1 = Categorical([unsorted_str, ints], ordered=True)
        c2 = Categorical([unsorted_str, ints], ordered=True)
        ac = Accum2(c2, c1)
        result = ac.sum(data)
        self.assertEqual(result._ncols, 8)
        for i, key1 in enumerate(sorted_str):
            k1 = bytes.decode(key1)
            k2 = sorted_ints[i]
            full_colname = "('" + k1 + "', " + str(k2) + ")"
            arr = result[full_colname]
            self.assertEqual(arr[i], sorted_data[i])

    def test_multikey_cats_filter_accum_sorted(self):
        unsorted_str = FastArray(["c", "e", "b", "d", "a"])
        ints = arange(1, 6) * 10
        data = np.random.rand(5) * 10

        # unsorted filter accum object
        c1 = Categorical([unsorted_str, ints])
        c2 = Categorical([unsorted_str, ints])
        ac = Accum2(c2, c1, showfilter=True)
        result = ac.sum(data)
        self.assertEqual(result._ncols, 9)
        for i, key1 in enumerate(unsorted_str):
            k1 = bytes.decode(key1)
            k2 = ints[i]
            full_colname = "('" + k1 + "', " + str(k2) + ")"
            arr = result[full_colname]
            self.assertEqual(arr[i + 1], data[i])

        # sorted filter accum object
        sortidx = np.argsort(unsorted_str)
        sorted_str = unsorted_str[sortidx]
        sorted_ints = ints[sortidx]
        sorted_data = data[sortidx]
        c1 = Categorical([unsorted_str, ints], sort_gb=True)
        c2 = Categorical([unsorted_str, ints], sort_gb=True)
        ac = Accum2(c2, c1, showfilter=True)
        result = ac.sum(data)
        self.assertEqual(result._ncols, 9)
        for i, key1 in enumerate(sorted_str):
            k1 = bytes.decode(key1)
            k2 = sorted_ints[i]
            full_colname = "('" + k1 + "', " + str(k2) + ")"
            arr = result[full_colname]
            # TODO fix this regression that was masked due to duplicate test names
            # self.assertAlmostEqual(arr[i + 1], sorted_data[i])

    def test_multikey_cats_filter_accum_ordered(self):
        unsorted_str = FastArray(["c", "e", "b", "d", "a"])
        ints = arange(1, 6) * 10
        data = np.random.rand(5) * 10

        # unsorted filter accum object
        c1 = Categorical([unsorted_str, ints])
        c2 = Categorical([unsorted_str, ints])
        ac = Accum2(c2, c1)
        result = ac.sum(data, showfilter=True)
        self.assertEqual(result._ncols, 9)
        for i, key1 in enumerate(unsorted_str):
            k1 = bytes.decode(key1)
            k2 = ints[i]
            full_colname = "('" + k1 + "', " + str(k2) + ")"
            arr = result[full_colname]
            self.assertEqual(arr[i + 1], data[i])

        # sorted filter accum object
        sortidx = np.argsort(unsorted_str)
        sorted_str = unsorted_str[sortidx]
        sorted_ints = ints[sortidx]
        sorted_data = data[sortidx]
        c1 = Categorical([unsorted_str, ints], ordered=True)
        c2 = Categorical([unsorted_str, ints], ordered=True)
        ac = Accum2(c2, c1)
        result = ac.sum(data, showfilter=True)
        self.assertEqual(result._ncols, 9)
        for i, key1 in enumerate(sorted_str):
            k1 = bytes.decode(key1)
            k2 = sorted_ints[i]
            full_colname = "('" + k1 + "', " + str(k2) + ")"
            arr = result[full_colname]
            self.assertEqual(arr[i + 1], sorted_data[i])

    def test_dataset_accum2(self):
        # test from accum2 off dataset and with a filter
        ds = Dataset({"test": arange(10), "data": arange(10) // 2})
        x = ds.accum2("data", "test").sum(ds.test, filter=ds.data == 3)
        totalcol = x.summary_get_names()[0]
        self.assertEqual(x[totalcol][3], 13)

    def test_accum2_mean(self):
        ds = Dataset({"time": arange(200.0)})
        ds.data = np.random.randint(7, size=200)
        ds.data2 = np.random.randint(7, size=200)
        symbols = ["AAPL", "AMZN", "FB", "GOOG", "IBM"]
        ds.symbol = Cat(1 + arange(200) % 5, symbols)
        ac = Accum2(ds.data, ds.symbol).mean(ds.time)
        totalcol = ac[ac.summary_get_names()[0]]
        footer = ac.footer_get_values()["Mean"]
        for i in range(len(symbols)):
            s_mean = ds[ds.symbol == symbols[i], :].time.mean()
            self.assertEqual(footer[i + 1], s_mean)
        for i in range(7):
            s_mean = ds[ds.data == i, :].time.mean()
            self.assertEqual(totalcol[i], s_mean)

    def test_accum2_median(self):
        ds = Dataset({"time": arange(200.0)})
        ds.data = np.random.randint(7, size=200)
        ds.data2 = np.random.randint(7, size=200)
        symbols = ["AAPL", "AMZN", "FB", "GOOG", "IBM"]
        ds.symbol = Cat(1 + arange(200) % 5, symbols)
        ac = Accum2(ds.data, ds.symbol).median(ds.time)
        totalcol = ac[ac.summary_get_names()[0]]
        footer = ac.footer_get_values()["Median"]
        for i in range(len(symbols)):
            s_median = ds[ds.symbol == symbols[i], :].time.median()
            self.assertEqual(footer[i + 1], s_median)
        for i in range(7):
            s_median = ds[ds.data == i, :].time.median()
            self.assertEqual(totalcol[i], s_median)

    def test_accum2_nanmedian_with_filter(self):
        ds = Dataset({"time": arange(200.0)})
        ds.data = np.random.randint(7, size=200)
        ds.data2 = np.random.randint(7, size=200)
        symbols = ["AAPL", "AMZN", "FB", "GOOG", "IBM"]
        # N.B. make a copy here for testing
        symbol_categorical = Cat(1 + arange(200) % 5, symbols)
        # N.B. Categorical.copy and Categorical constructor doesn't do deep copy?!
        ds.symbol = Cat(1 + arange(200) % 5, symbols)

        chosen_symbols = ["AMZN", "AAPL"]
        filt = symbol_categorical.isin(chosen_symbols)
        ac = Accum2(ds.data, ds.symbol)
        stat1 = ac.nanmedian(ds.time, filter=filt)
        totalcol = stat1[stat1.summary_get_names()[0]]
        footer = stat1.footer_get_values()["Median"]
        # Make sure we don't change the input data
        self.assertTrue(not rt.any(ds.symbol._fa == 0))
        for sym in chosen_symbols:
            s_median = rt.nanmedian(ds[symbol_categorical == sym, :].time)
            i = rt.where(symbol_categorical.category_array == sym)[0].item()
            self.assertEqual(footer[i + 1], s_median)
        for i in range(7):
            s_median = rt.nanmedian(ds[(ds.data == i) & filt, :].time)
            self.assertEqual(totalcol[i], s_median)

        chosen_symbols = ["IBM", "FB"]
        filt = symbol_categorical.isin(chosen_symbols)
        stat2 = ac.nanmedian(ds.time, filter=filt)
        totalcol = stat2[stat2.summary_get_names()[0]]
        footer = stat2.footer_get_values()["Median"]
        # Make sure we don't change the input data
        self.assertTrue(not rt.any(ds.symbol._fa == 0))
        for sym in chosen_symbols:
            s_median = rt.nanmedian(ds[symbol_categorical == sym, :].time)
            i = rt.where(symbol_categorical.category_array == sym)[0].item()
            self.assertEqual(footer[i + 1], s_median)
        for i in range(7):
            s_median = rt.nanmedian(ds[(ds.data == i) & filt, :].time)
            self.assertEqual(totalcol[i], s_median)

    def test_accum2_quantile(self):
        kwargs = {QUANTILE_METHOD_NP_KW: "midpoint"}
        symbols = ["AAPL", "AMZN", "FB", "GOOG", "IBM"]

        def footer_name(q):
            if q == 0.5:
                return "Median"
            elif q == 0.0:
                return "Min"
            elif q == 1.0:
                return "Max"
            else:
                return "Quantile"

        # large N to check multithreading
        for N in [2345, 123_985]:
            ds = Dataset({"time": arange(N)})
            ds.symbol = Cat(1 + arange(N) % 5, symbols)
            for _ in range(2):
                ds.data = np.random.randint(-15, 15, size=N)
                ds.data2 = np.random.randn(N)
                for q in [0.0, 0.154, 0.482, 0.5, 0.91, 1.0]:
                    ac = Accum2(ds.data, ds.symbol).quantile(ds.time, q=q)
                    totalcol = ac[ac.summary_get_names()[0]]
                    footer = ac.footer_get_values()[footer_name(q)]
                    for i in range(len(symbols)):
                        s_quantile = ds[ds.symbol == symbols[i], :].time.quantile(q=q, **kwargs)
                        self.assertEqual(footer[i + 1], s_quantile)
                    for val in range(-15, 15):
                        curr_slice = ds[ds.data == val, :].time
                        if not len(curr_slice):
                            continue
                        idx = np.arange(len(ac))[ac.data == val]
                        s_quantile = curr_slice.quantile(q=q, **kwargs)
                        self.assertEqual(totalcol[idx], s_quantile)

                    # quantile and percentile should be the same
                    ac_p = Accum2(ds.data, ds.symbol).percentile((ds.time, ds.data2), q=(q * 100))
                    ac_q = Accum2(ds.data, ds.symbol).quantile((ds.time, ds.data2), q=q)
                    for multiset_key in ac_p.keys():
                        for col in ac_p[multiset_key].keys():
                            assert_array_almost_equal(ac_p[multiset_key][col], ac_q[multiset_key][col])

    def test_accum2_nanquantile_with_filter(self):
        kwargs = {QUANTILE_METHOD_NP_KW: "midpoint"}
        symbols = ["AAPL", "AMZN", "FB", "GOOG", "IBM"]

        def footer_name(q):
            if q == 0.5:
                return "Median"
            elif q == 0.0:
                return "Nanmin"
            elif q == 1.0:
                return "Nanmax"
            else:
                return "Nanquantile"

        # large N to check multithreading
        for N in [1_045, 78_013]:
            # TODO: fix this for integer types. Min (nanmin) puts `Inv` as the smallest value
            ds = Dataset({"time": arange(N).astype(float)})
            ds.symbol = Cat(1 + arange(N) % 5, symbols)
            for _ in range(2):
                ds.data = np.random.randint(-10, 20, size=N)
                ds.data2 = np.random.random(size=N)
                # N.B. make a copy here for testing
                symbol_categorical = Cat(1 + arange(N) % 5, symbols)
                # N.B. Categorical.copy and Categorical constructor doesn't do deep copy?!
                ds.symbol = Cat(1 + arange(N) % 5, symbols)
                for q in [0.0, 0.001, 0.333, 0.5, 0.67, 0.831, 1.0]:
                    chosen_symbols = ["AMZN", "AAPL"]
                    filt = symbol_categorical.isin(chosen_symbols)
                    ac = Accum2(ds.data, ds.symbol)
                    stat1 = ac.nanquantile(ds.time, q=q, filter=filt)
                    totalcol = stat1[stat1.summary_get_names()[0]]
                    footer = stat1.footer_get_values()[footer_name(q)]
                    # Make sure we don't change the input data
                    self.assertTrue(not rt.any(ds.symbol._fa == 0))

                    for sym in chosen_symbols:
                        s_quantile = ds[symbol_categorical == sym, :].time.nanquantile(q=q, **kwargs)
                        i = rt.where(symbol_categorical.category_array == sym)[0].item()
                        self.assertEqual(footer[i + 1], s_quantile)
                    for val in range(-10, 20):
                        curr_slice = ds[(ds.data == val) & filt, :].time
                        if not len(curr_slice):
                            continue
                        idx = np.arange(len(stat1))[stat1.data == val]
                        s_quantile = curr_slice.nanquantile(q=q, **kwargs)
                        self.assertEqual(totalcol[idx], s_quantile)

                    chosen_symbols = ["IBM", "FB"]
                    filt = symbol_categorical.isin(chosen_symbols)
                    stat2 = ac.nanquantile(ds.time, q=q, filter=filt)
                    totalcol = stat2[stat2.summary_get_names()[0]]
                    footer = stat2.footer_get_values()[footer_name(q)]
                    # Make sure we don't change the input data
                    self.assertTrue(not rt.any(ds.symbol._fa == 0))
                    for sym in chosen_symbols:
                        s_quantile = ds[symbol_categorical == sym, :].time.nanquantile(q=q, **kwargs)
                        i = rt.where(symbol_categorical.category_array == sym)[0].item()
                        self.assertEqual(footer[i + 1], s_quantile)
                    for val in range(-10, 20):
                        curr_slice = ds[(ds.data == val) & filt, :].time
                        if not len(curr_slice):
                            continue
                        idx = np.arange(len(stat2))[stat2.data == val]
                        s_quantile = curr_slice.nanquantile(q=q, **kwargs)
                        self.assertEqual(totalcol[idx], s_quantile)

    def test_showfilter_label_subclass(self):
        d = Date.range("20190201", "20190210")
        c = Categorical(d)
        c2 = Categorical(arange(10))
        ac = Accum2(c, c2)
        result = ac.count(showfilter=True)

        self.assertTrue(isinstance(result.YLabel, Date))
        self.assertTrue(result.YLabel.isnan()[0])

        d = DateTimeNano.random(10)
        c = Categorical(d)
        c2 = Categorical(arange(10))
        ac = Accum2(c, c2)
        result = ac.count(showfilter=True)

        self.assertTrue(isinstance(result.YLabel, DateTimeNano))
        self.assertTrue(result.YLabel.isnan()[0])

        d = DateSpan(arange(10, 20))
        c = Categorical(d)
        c2 = Categorical(arange(10))
        ac = Accum2(c, c2)
        result = ac.count(showfilter=True)

        self.assertTrue(isinstance(result.YLabel, DateSpan))
        self.assertTrue(result.YLabel.isnan()[0])

        d = TimeSpan(np.random.rand(10) * 10_000_000_000)
        c = Categorical(d)
        c2 = Categorical(arange(10))
        ac = Accum2(c, c2)
        result = ac.count(showfilter=True)

        self.assertTrue(isinstance(result.YLabel, TimeSpan))
        self.assertTrue(result.YLabel.isnan()[0])

    def test_apply(self):
        arrsize = 200
        numrows = 7

        ds = Dataset({"time": arange(arrsize * 1.0)})
        ds.data = np.random.randint(numrows, size=arrsize)
        ds.data2 = np.random.randint(numrows, size=arrsize)
        symbols = ["AAPL", "AMZN", "FB", "GOOG", "IBM"]
        ds.symbol = Cat(1 + arange(arrsize) % len(symbols), symbols)
        ds.accum2("symbol", "data").sum(ds.data2)
        ds.accum2("symbol", "data").sum(ds.data2, showfilter=True)
        ds.accum2("symbol", "data").median(ds.data2, showfilter=True)
        ds.accum2("symbol", "data").median(ds.data2, showfilter=False)
        ds.accum2("symbol", "data").quantile(ds.data2, showfilter=True, q=0.35)
        ds.accum2("symbol", "data").quantile(ds.data2, showfilter=False, q=0.75)
        ds.accum2("symbol", "data").nanquantile(ds.data2, showfilter=True, q=[0.0, 0.54, 0.999])
        ds.accum2("symbol", "data").nanquantile(ds.data2, showfilter=False, q=[0.000001, 0.5, 0.31])
        ds.accum2("symbol", "data").percentile(ds.data2, showfilter=True, q=[10, 20, 100])
        ds.accum2("symbol", "data").nanpercentile(ds.data2, showfilter=False, q=0.000001)
        ds.accum2("symbol", "data").apply_reduce(np.median, ds.data2, showfilter=True)
        ds.accum2("symbol", "data").apply_reduce(np.median, ds.data2, showfilter=False)
        f = logical(arange(200) % 2)
        ds.accum2("symbol", "data").apply_reduce(np.median, ds.data2, filter=f)
        ds.accum2("symbol", "data").apply_reduce(np.median, ds.data2, filter=f, showfilter=True)
        ds.accum2("symbol", "data").median(ds.data2, filter=f, showfilter=True)

    def test_apply_nonreduce(self):
        arrsize = 200
        numrows = 7
        ds = rt.Dataset({"time": rt.arange(arrsize * 1.0)})
        ds.data = arange(arrsize) % numrows
        ds.data2 = (arange(arrsize) + 3) % numrows
        symbols = [
            "AAPL",
            "AMZN",
            "FB",
            "GOOG",
            "IBM",
            "6",
            "7",
            "8",
            "9",
            "10",
            "11",
            "12",
            "13",
            "14",
            "15",
            "16",
            "17",
            "18",
        ]
        ds.symbol = rt.Cat(1 + rt.arange(arrsize) % len(symbols), symbols)
        result = ds.symbol.apply_reduce(lambda x, y: np.sum(np.minimum(x, y)), (ds.data, ds.data))

        ac = ds.accum2("symbol", "data")
        newds = ac.apply_nonreduce(np.cumsum)
        ds2 = ac.apply_reduce(lambda x, y: np.sum(np.maximum(x, y)), (newds.data, newds.data2))

        x = np.maximum(newds.data, newds.data2)
        y = ac.apply_nonreduce(lambda x, y: np.maximum(x, y), (newds.data, newds.data2))[0]
        self.assertTrue(np.all(x == y))


class AccumTable_Test(unittest.TestCase):
    @pytest.mark.skip(
        reason="Test needs to be re-written to remove the np.random.seed usage -- it's not stable across numpy versions."
    )
    def test_accum_table(self):

        # Create the test data

        def unpivot(frame):
            N, K = frame.shape
            data = {
                "value": frame.values.ravel("F"),
                "variable": np.asarray(frame.columns).repeat(N),
                "date": np.tile(np.asarray(frame.index), K),
            }
            return pd.DataFrame(data, columns=["date", "variable", "value"])

        np.random.seed(1234)
        df = unpivot(pd.concat([tm.makeTimeDataFrame(), tm.makeTimeDataFrame()]))
        ds = dataset_from_pandas_df(df)
        ds.date = DateTimeNano(ds.date, from_tz="NYC").to_iso()
        ds.date = rt.FastArray([d[:10] for d in ds.date])
        ds.variable = rt.Categorical(ds.variable)
        ds.date = rt.Categorical(ds.date)

        at = rt.AccumTable(ds.date, ds.variable)

        # Add and view inner tables with totals
        at["Sum"] = at.sum(ds.value)
        self.assertEqual(at["Sum"].shape, (3, 7))
        assert_array_almost_equal(at["Sum"]["A"], np.array([0.47, -0.79, 1.72]), decimal=2)

        vw = at.gen("Sum")
        self.assertEqual(vw.shape, (3, 7))
        assert_array_almost_equal(vw["A"], np.array([0.47, -0.79, 1.72]), decimal=2)

        assert_array_almost_equal(vw["Sum"], np.array([-0.10, -5.02, 5.37]), decimal=2)
        assert_array_almost_equal(vw.footer_get_values(columns=["Sum"])["Sum"], np.array([0.25]), decimal=2)

        at["Mean"] = at.mean(ds.value)
        self.assertEqual(at["Mean"].shape, (3, 7))
        assert_array_almost_equal(at["Mean"]["A"], np.array([0.24, -0.39, 0.86]), decimal=2)

        at["Half"] = at["Mean"] / at["Sum"]
        self.assertEqual(at["Half"].shape, (3, 7))
        assert_array_almost_equal(at["Half"]["A"], np.array([0.5, 0.5, 0.5]), decimal=2)

        # Add and view inner tables with blanks

        at["Blanks"] = at["Sum"].copy()
        at["Blanks"]["C"] = 0.0
        for col in at["Blanks"][:, 1:]:
            at["Blanks"][col][2] = np.nan

        vw = at.gen("Blanks")
        self.assertEqual(vw.shape, (2, 9))
        assert_array_almost_equal(vw["A"], np.array([0.47, -0.79]), decimal=2)
        assert_array_almost_equal(vw["Blanks"], np.array([-0.10, -5.02]), decimal=2)
        self.assertAlmostEqual(vw.footer_get_dict()["Blanks"]["Blanks"], 0.245, places=2)

        vw = at.gen("Blanks", remove_blanks=False)
        self.assertEqual(vw.shape, (3, 10))
        assert_array_almost_equal(vw["A"], np.array([0.47, -0.79, np.nan]), decimal=2)
        assert_array_almost_equal(vw["Blanks"], np.array([-0.10, -5.02, np.nan]), decimal=2)

        # Test division with zeros and nans
        at["Bad"] = at["Blanks"] / at["Half"]
        self.assertEqual(at["Blanks"].shape, (3, 7))
        vw = at.gen("Bad")
        self.assertEqual(vw.shape, (2, 10))
        vw = at.gen("Blanks")
        self.assertEqual(vw.shape, (2, 10))
        vw = at.gen("Half")
        self.assertEqual(vw.shape, (3, 11))

        # Set margin columns to the right

        at.set_margin_columns(["Blanks", "Mean"])
        vw = at.gen("Half")
        self.assertEqual(vw.shape, (3, 9))
        self.assertEqual(vw.keys()[6], "Half")
        self.assertEqual(vw.keys()[7], "Blanks")
        self.assertEqual(vw.keys()[8], "Mean")
        self.assertEqual(list(vw.footer_get_dict().keys()), ["Half", "Sum", "Mean", "Blanks", "Bad"])

        vw = at.gen()
        self.assertEqual(vw.keys()[6], "Half")

        vw = at.gen("Sum")
        self.assertEqual(vw.keys()[6], "Sum")
        self.assertEqual(vw.keys()[7], "Blanks")
        self.assertEqual(vw.keys()[8], "Mean")
        self.assertEqual(list(vw.footer_get_dict().keys()), ["Sum", "Mean", "Half", "Blanks", "Bad"])

        # Set footer rows at the bottom

        at.set_footer_rows(["Mean"])
        vw = at.gen("Half")
        self.assertEqual(vw.shape, (3, 9))
        self.assertEqual(vw.keys()[6], "Half")
        self.assertEqual(vw.keys()[7], "Blanks")
        self.assertEqual(vw.keys()[8], "Mean")
        self.assertEqual(list(vw.footer_get_dict().keys()), ["Half", "Mean"])

        vw = at.gen("Sum")
        self.assertEqual(vw.keys()[6], "Sum")
        self.assertEqual(vw.keys()[7], "Blanks")
        self.assertEqual(vw.keys()[8], "Mean")
        self.assertEqual(list(vw.footer_get_dict().keys()), ["Sum", "Mean"])

        # Access view Dataset elements

        vw = at.gen("Sum")
        assert_array_equal(vw.date, rt.FastArray(["2000-01-03", "2000-01-04", "2000-01-05"]))
        assert_array_almost_equal(vw["Sum"], np.array([-0.10, -5.02, 5.37]), decimal=2)
        assert_almost_equal(vw[vw.date == "2000-01-03", "A"][0], 0.47355353, decimal=2)
        assert_almost_equal(
            list(vw.footer_get_values("Sum", columns=["A"]).values())[0],
            1.409830,
            decimal=2,
        )


if __name__ == "__main__":
    tester = unittest.main()



"""This module has side effects on import.
If DisplayOptions.CUSTOM_COMPLETION is True then this will disable IPython Jedi and
use a IPCompleter custom complete.
"""
import sys
from collections import ChainMap
from types import MethodType
from typing import Iterable, List, Optional, Tuple

from IPython import get_ipython
from IPython.core.completer import _FakeJediCompletion
from IPython.utils.generics import complete_object

import riptable

__all__ = ["enable_custom_attribute_completion"]


_PRESERVE_ORDER_SENTINEL: str = "PRESERVE_ORDER_SENTINEL"
_DEBUG: bool = False  # toggle to enable stdout logging


def _is_public(name: str) -> bool:
    """Returns True if a method is not protected, nor private; otherwise False."""
    return not (name.startswith("__") or name.startswith("_"))


def _get_key_names(obj: object) -> List[str]:
    if hasattr(obj, "keys"):
        return obj.keys()
    return []


def _ns_search(name: str) -> Optional[object]:
    """
    IPython specific search of local then global namespace for name.
    If found returns the object, otherwise return None.
    """
    from IPython import get_ipython

    ip = get_ipython()

    # check ipython namespace table
    chain_map = ChainMap(
        ip.ns_table.get("user_local"),
        ip.ns_table.get("user_global"),
        ip.ns_table.get("builtin"),
    )
    if name in chain_map:
        return chain_map.get(name)

    # check python namespace; if ipython namespace is additive to the python namespace then this is redundant
    if name in locals():
        return locals()[name]
    if name in globals():
        return globals()[name]

    return None


# The motivation for this monkey patched completion is to add support for custom completion that preserve ordering
# while remaining backwards compatible with the original IPython completion machinery (see RIP-111 for more details).
# IPython version 7.10.2supports two approaches for custom completion that didn't work for us:
# 1) IPython supports custom completer dispatching for python objects.
#   https://ipython.readthedocs.io/en/stable/api/generated/IPython.utils.generics.html#IPython.utils.generics.complete_object
# 2) IPython supports custom completion functions
#   https://ipython.readthedocs.io/en/stable/api/generated/IPython.core.interactiveshell.html#IPython.core.interactiveshell.InteractiveShell.set_custom_completer
# Approach (1) does post filtering and ordering after the custom completer has run.
# Approach (2) - set_custom_completer attempts to insert the completer in the IPCompleter.matchers property which
# returns a list literal thereby dropping the custom completer insertion.
def _monkey_patched__complete(
    self, *, cursor_line, cursor_pos, line_buffer=None, text=None, full_text=None
) -> Tuple[str, List[str], List[str], Iterable[_FakeJediCompletion]]:
    """

    Like complete but can also returns raw jedi completions as well as the
    origin of the completion text. This could (and should) be made much
    cleaner but that will be simpler once we drop the old (and stateful)
    :any:`complete` API.


    With current provisional API, cursor_pos act both (depending on the
    caller) as the offset in the ``text`` or ``line_buffer``, or as the
    ``column`` when passing multiline strings this could/should be renamed
    but would add extra noise.
    """

    # if the cursor position isn't given, the only sane assumption we can
    # make is that it's at theendregion of the line (the common case)
    if cursor_pos is None:
        cursor_pos = len(line_buffer) if text is None else len(text)

    if self.use_main_ns:
        self.namespace = __main__.__dict__  # noqa: F821

    # if text is either None or an empty string, rely on the line buffer
    if (not line_buffer) and full_text:
        line_buffer = full_text.split("\n")[cursor_line]
    if not text:
        text = self.splitter.split_line(line_buffer, cursor_pos)

    if self.backslash_combining_completions:
        # allow deactivation of these on windows.
        base_text = text if not line_buffer else line_buffer[:cursor_pos]
        latex_text, latex_matches = self.latex_matches(base_text)
        if latex_matches:
            return latex_text, latex_matches, ["latex_matches"] * len(latex_matches), ()
        name_text = ""
        name_matches = []

        # region monkey patched region
        # The latex namespace searches can be more defensive by falling back to a local implementation of these function
        # definitions if they are not found from _ns_search.
        back_latex_name_matches, back_unicode_name_matches = (
            _ns_search("back_latex_name_matches"),
            _ns_search("back_unicode_name_matches"),
        )
        if back_latex_name_matches and back_unicode_name_matches:
            # endregion monkey patched region
            # need to add self.fwd_unicode_match() function here when done
            for meth in (
                self.unicode_name_matches,
                back_latex_name_matches,
                back_unicode_name_matches,
                self.fwd_unicode_match,
            ):
                name_text, name_matches = meth(base_text)
                if name_text:
                    # region monkey patched region
                    # As of IPython 7.10.2, MATCHES_LIMIT is set to 500
                    MATCHES_LIMIT = _ns_search("MATCHES_LIMIT") if _ns_search("MATCHES_LIMIT") else 500
                    # endregion monkey patched region
                    return (
                        name_text,
                        name_matches[:MATCHES_LIMIT],
                        [meth.__qualname__] * min(len(name_matches), MATCHES_LIMIT),
                        (),
                    )

    # If no line buffer is given, assume the input text is all there was
    if line_buffer is None:
        line_buffer = text

    self.line_buffer = line_buffer
    self.text_until_cursor = self.line_buffer[:cursor_pos]

    # Do magic arg matches
    MATCHES_LIMIT = 500  # get this from local
    for matcher in self.magic_arg_matchers:
        matches = list(matcher(line_buffer))[:MATCHES_LIMIT]
        if matches:
            origins = [matcher.__qualname__] * len(matches)
            return text, matches, origins, ()

    # Start with a clean slate of completions
    matches = []

    # FIXME: we should extend our api to return a dict with completions for
    # different types of objects.  The rlcomplete() method could then
    # simply collapse the dict into a list for readline, but we'd have
    # richer completion semantics in other environments.
    completions = ()
    if self.use_jedi:
        if not full_text:
            full_text = line_buffer
        completions = self._jedi_matches(cursor_pos, cursor_line, full_text)

    if self.merge_completions:
        matches = []
        for matcher in self.matchers:
            try:
                matches.extend([(m, matcher.__qualname__) for m in matcher(text)])
            except:
                # Show the ugly traceback if the matcher causes an
                # exception, but do NOT crash the kernel!
                sys.excepthook(*sys.exc_info())
    else:
        for matcher in self.matchers:
            matches = [(m, matcher.__qualname__) for m in matcher(text)]
            if matches:
                break

    seen = set()
    filtered_matches = set()
    for m in matches:
        t, c = m
        if t not in seen:
            filtered_matches.add(m)
            seen.add(t)

    _filtered_matches = []
    if _DEBUG:
        print(
            f"IPCompleter.monkey_patched__complete: check if {text + _PRESERVE_ORDER_SENTINEL} in matches={[m[0] for m in matches]}\n"
        )
    # region monkey patched region
    matches_text = [
        m[0] for m in matches
    ]  # unpack the text part of the list of tuples of text matches and matcher name
    if len(matches) and text + _PRESERVE_ORDER_SENTINEL in matches_text:
        matches.pop()  # remove the PRESERVE_ORDER_SENTINEL
        _filtered_matches = matches
        if _DEBUG:
            print(
                f"IPCompleter.monkey_patched__complete: custom completion path: check matches={[m[0] for m in matches]}\n\tmatches{matches}"
            )
    else:
        completions_sorting_key_fn = _ns_search("completions_sorting_key")
        if not completions_sorting_key_fn:
            # completions_sorting_key is pulled in from IPython 7.10.2 in the event we cannot find this function in the namespace
            def completions_sorting_key(word):
                """key for sorting completions

                This does several things:

                - Demote any completions starting with underscores to the end
                - Insert any %magic and %%cellmagic completions in the alphabetical order
                  by their name
                """
                prio1, prio2 = 0, 0

                if word.startswith("__"):
                    prio1 = 2
                elif word.startswith("_"):
                    prio1 = 1

                if word.endswith("="):
                    prio1 = -1

                if word.startswith("%%"):
                    # If there's another % in there, this is something else, so leave it alone
                    if not "%" in word[2:]:
                        word = word[2:]
                        prio2 = 2
                elif word.startswith("%"):
                    if not "%" in word[1:]:
                        word = word[1:]
                        prio2 = 1

                return prio1, word, prio2

            completions_sorting_key_fn = completions_sorting_key
        _filtered_matches = sorted(filtered_matches, key=lambda x: completions_sorting_key_fn(x[0]))
    # endregion monkey patched region

    custom_res = [(m, "custom") for m in self.dispatch_custom_completer(text) or []]

    _filtered_matches = custom_res or _filtered_matches

    _filtered_matches = _filtered_matches[:MATCHES_LIMIT]
    _matches = [m[0] for m in _filtered_matches]
    origins = [m[1] for m in _filtered_matches]

    self.matches = _matches

    return text, _matches, origins, completions


def _disable_jedi() -> None:
    ip = get_ipython()
    ip.Completer.use_jedi = False


def _set__monkey_patched__complete() -> None:
    ip = get_ipython()
    ip.Completer._complete = MethodType(_monkey_patched__complete, ip.Completer)


def _set_custom_attribute_completion() -> None:
    @complete_object.register(object)
    def _attribute_method_completion(obj: object, existing_completions: List[str]) -> List[str]:
        key_names: List[str] = []  # grab keys if they are available
        if hasattr(obj, "_ipython_key_completions_"):
            key_names = obj._ipython_key_completions_()
        elif hasattr(obj, "keys"):
            key_names = obj.keys()

        make_sorted_list = lambda lst: sorted(list(lst), key=str.casefold)  # Type Callable[[Iterable], List]

        # Return the sorted keys, then sorted existing completions, then a sentinel
        # for the custom completion to signal preserving this ordering.
        key_names = make_sorted_list(key_names)
        existing_public_completions = make_sorted_list(filter(_is_public, existing_completions))
        ordered_completions = {c: None for c in key_names + existing_public_completions}
        matches = list(ordered_completions.keys())
        matches.append(_PRESERVE_ORDER_SENTINEL)
        return matches


# As of 20191218, if `DisplayOptions.CUSTOM_COMPLETION` is toggled on it results in a one-time registration of custom
# attribute completion per IPython session as opposed to supporting deregistration in the same session.
# The following is a list of tasks to make this more extensible and robust:
# - Create a CustomCompleterManager that manages registration and deregistration more cleanly.
# - Investigate using traitlets observers for on the fly toggling.
# - Investigate deregistration of custom completer.
# - Add configuration options to allow for other data structures such pandas DataFrames custom attribute
#   completion ordering.
def enable_custom_attribute_completion() -> None:
    """
    Registers the custom attribute completer and sets our monkey patched complete method to preserve the custom
    attribute completer ordering.
    If an object has keys then the completion results will show the keys then any existing completions
    in case insensitive sorted order.

    Notable side effect - this will disable Jedi support since IPCompleter will
    prefer to use Jedi completions over any custom completer.
    """

    _disable_jedi()

    _set__monkey_patched__complete()

    _set_custom_attribute_completion()


# Note, this module has side effects!
if riptable.TypeRegister.DisplayOptions.CUSTOM_COMPLETION:
    enable_custom_attribute_completion()



__all__ = ["hstack_any", "stack_rows"]

import warnings
from collections import defaultdict
from typing import List, Mapping, Optional, Union

import numpy as np
import riptide_cpp as rc

from .rt_enum import CategoryMode, NumpyCharTypes, TypeRegister, int_dtype_from_len
from .rt_grouping import hstack_groupings, merge_cats
from .rt_numpy import arange, empty, hstack


def hstack_any(
    itemlist: Union[list, Mapping[str, np.ndarray]],
    cls: Optional[type] = None,
    baseclass: Optional[type] = None,
    destroy: bool = False,
    **kwargs,
):
    """
    stack_rows or hstack_any is the main routine to row stack riptable classes.
    It stacks categoricals, datasets, time objects, structs.
    It can now stack a dictionary of numpy arrays to return a single array and a categorical.

    Parameters
    ----------
    itemlist:
        a list of objects to stack for arrays, datasets, categoricals
        a dictionary of numpy arrays

    Other Parameters
    ----------------
    cls: None.  the type of class we are stacking
    baseclass: the baseclass we are stacking
    destroy: bool, False.  Only valid for Datasets
        !! This is dangerous so make sure you do not want the data anymore in the original datasets.

    Returns
    -------
    In the case of a list: returns a single new array, dataset, categorical or specified object.
    In the case of a dict: returns a single new array and a new categorical (two objects returned).

    Examples
    --------
    >>> stack_rows([arange(3), arange(2)])
    FastArray([0, 1, 2, 0, 1])

    >>> d={'test1':arange(3), 'test2':arange(1), 'test3':arange(2)}
    >>> arr, cat = stack_rows(d)
    >>> Dataset({'Data':arr, 'Cat': cat})
    #   Data   Cat
    -   ----   -----
    0      0   test1
    1      1   test1
    2      2   test1
    3      0   test2
    4      0   test3
    5      1   test3

    See Also
    --------
    np.hstack
    rt.Categorical.align
    rt.Dataset.concat_rows
    """
    # ------------------------------
    def hstack_containers(itemlist, fill_inv=True, destroy=False):
        # assembles all columns into lists for final hstack call
        # keeps running count for gap, creates invalid fill array
        # the invalid fill should get pushed down, provide cutoffs instead
        # should there be a flag to fill with invalid? can't get row length from structs

        # final call to hstack or vstack
        # hstack will call subclass hstack

        # keep track of item order as more are added
        # use the first container to initialize order
        id = len(itemlist[0])
        allkeys = dict(zip(itemlist[0], arange(id)))
        alldicts = {
            "Array": defaultdict(list),
            "Struct": defaultdict(list),
            "Dataset": defaultdict(list),
            "PDataset": defaultdict(list),
        }

        newitems = []
        for container in itemlist:
            for key, item in container.items():
                exists = allkeys.setdefault(key, id)
                if exists == id:
                    newitems.append(key)
                id += 1

                # wrap scalars
                if np.isscalar(item):
                    item = TypeRegister.FastArray(item)

                if isinstance(item, np.ndarray):
                    dname = "Array"

                elif issubclass(type(item), TypeRegister.Struct):
                    dname = type(item).__name__

                else:
                    raise TypeError(f"Unsupported type {type(item)} in container hstack.")

                # check for different types in different containers
                for dkey, d in alldicts.items():
                    if dkey == dname:
                        continue
                    if key in d:
                        # maybe raise a value error?
                        warnings.warn(f"Found conflicting types for item {key} {dkey} vs. {dname}. Data will be lost.")

                # add to list of items for that type
                alldicts[dname][key].append(item)

        # warn on missing items
        if len(newitems) > 0:
            warnings.warn(f"These items were not found in every container: {newitems}")

        # array will fix subclass
        for k, v in alldicts["Array"].items():
            alldicts["Array"][k] = hstack_arrays(v)

        # for all container types (possibly allow these to get stacked together?)
        for k, v in alldicts["Struct"].items():
            alldicts["Struct"][k] = hstack_containers(v)

        for k, v in alldicts["Dataset"].items():
            alldicts["Dataset"][k] = hstack_containers(v, destroy=destroy)

        for k, v in alldicts["PDataset"].items():
            alldicts["PDataset"][k] = hstack_containers(v)

        # combine dicts, allkeys is final sort order
        result = {}
        for d in alldicts.values():
            result.update(d)
        return result, list(allkeys)

    # ------------------------------
    def hstack_arrays(itemlist):
        # maybe pass cutoffs to hstack arrays for invalid fill?

        # for all array-like items:
        # check type
        # check number of dimensions
        # check shape

        # for dtypes that don't match, check in python or send down?
        # right now:
        # checks that all are a numeric type or all are a string type, warns
        # cannot stack FA subclass with regular FA
        return hstack(itemlist)

    # ------------------------------
    def hstack_dict(itemdict, **kwargs):
        r"""
        Will create a Categorical from the keys in the dictionary.

        Examples
        --------
        >>> mydict={}
        >>> basepath = r'C:\path\to\data'
        >>> for j in range(11):
        ...    filename = f'{basepath}{j}.sds'
        ...    ds = rt.load_sds(filename)
        ...    mydict[f'cat{j}']=ds
        >>> z, mycat = rt.stack_rows(mydict)
        >>> z.mycat
        """
        dictlen = len(itemdict)
        if dictlen == 0:
            raise ValueError(f"List of dictionary items to stack was empty.")

        # create an array of cutoffs for slicing later
        lengths = TypeRegister.FastArray([z.shape[0] for z in itemdict.values()], dtype=np.int64).cumsum()
        keynames = [*itemdict.keys()]

        return_arr = hstack([*itemdict.values()])

        # choose proper ikey size
        total_len = return_arr.shape[0]

        if total_len != lengths[-1]:
            raise ValueError(
                f"Final lengths do not match.  The hstack resulted in length {total_len} but counted a total length of {lengths[-1]}."
            )

        ikey = empty((total_len,), dtype=int_dtype_from_len(total_len))

        startpos = 0
        for i in range(len(lengths)):
            endpos = lengths[i]
            ikey[startpos:endpos] = i + 1
            startpos = endpos

        return_cat = TypeRegister.Categorical(ikey, keynames, ordered=False)
        return return_arr, return_cat

    # start of hstack_any ---------
    # ------------------------------
    # map each riptable type to a custom hstack function
    hstack_funcs = {
        TypeRegister.Date: _hstack_date,
        TypeRegister.DateSpan: _hstack_datespan,
        TypeRegister.DateTimeNano: _hstack_datetimenano,
        TypeRegister.TimeSpan: _hstack_timespan,
        TypeRegister.Categorical: _hstack_categorical,
        TypeRegister.Dataset: _hstack_dataset,
        TypeRegister.Struct: _hstack_struct,
    }

    # items need to come in a tuple or list or dict
    if isinstance(itemlist, dict):
        return hstack_dict(itemlist, **kwargs)

    if not isinstance(itemlist, (list, tuple)):
        raise TypeError(f"Input must be a list or tuple of datasets. Got {type(itemlist)}")

    if len(itemlist) == 1:
        return itemlist[0]
    elif len(itemlist) == 0:
        raise ValueError(f"List of items to stack was empty.")

    sameclass = {type(o) for o in itemlist}
    if len(sameclass) != 1:
        error = True
        if len(sameclass) == 2:
            tempset = sameclass.copy()
            t1 = tempset.pop()
            t2 = tempset.pop()

            # If one is base class of the other, use that.
            # happens with Dataset and PDataset
            if issubclass(t1, t2):
                sameclass = {t2}
                error = False
            elif issubclass(t2, t1):
                sameclass = {t1}
                error = False
        if error:
            raise TypeError(f"itemlist must contain objects with the same class: {sameclass!r}")

    if cls is None:
        cls = sameclass.pop()

    # hit the old routines
    if cls is not None:
        func = hstack_funcs.get(cls)
        if func is not None:
            return func(itemlist, destroy=destroy)

    # fallback to baseclass
    if baseclass is not None:
        func = hstack_funcs.get(baseclass)
        if func is not None:
            return func(itemlist, destroy=destroy)

    # default to normal hstack
    return hstack_arrays(itemlist)


# --------------------------------------------------------------------------
def _hstack_struct(struct_list, destroy: bool = False):
    """
    Merges data from multiple structs.

    A struct utility for merging data from multiple structs (useful for multiday loading).
    Structs must have the same keys, and contain only Structs, Datasets, arrays, and riptable arrays.

    Parameters
    ----------
    struct_list : obj:`list` of obj:`Struct`
        A list of Struct instances to hstack.
    destroy : bool
        This parameter is currently ignored.

    Returns
    -------
    obj:`Struct`

    See Also
    --------
    riptable.hstack
    """
    # items will be sorted into lists by type for final hstack call
    alldicts = defaultdict(dict)

    if len(struct_list) == 1 and isinstance(struct_list[0], TypeRegister.Struct):
        return struct_list[0]
    elif len(struct_list) > 0:
        # assumes that all keys are in the first struct
        # model_keys = struct_list[0].keys()
        # initialize keys with those from the first struct
        id = len(struct_list[0])
        all_keys = dict(zip(struct_list[0].keys(), arange(id)))
    else:
        raise ValueError(f"List of structs was empty.")

    new_items = []
    for st in struct_list:
        for key, item in st.items():
            # check if item has already been marked for concatenation
            exists = all_keys.setdefault(key, id)
            if exists == id:
                # this is a new item that did not appear in the structs before it
                # warn later
                new_items.append(key)
            id += 1

            # might change this in the future. matlab scalars now always load as arrays.
            if np.isscalar(item):
                item = TypeRegister.FastArray(item)

            # group together riptable types
            if isinstance(item, TypeRegister.Struct) or TypeRegister.is_array_subclass(item):
                dname = type(item).__name__

            # for all regular numpy / fastarray
            elif isinstance(item, np.ndarray):
                dname = "Array"

            else:
                raise TypeError(f"Unsupported type {type(item)} in Struct append.")

            # check for key in other type dictionaries
            for dkey, d in alldicts.items():
                if dkey == dname:
                    continue
                if key in d:
                    warnings.warn(f"Found conflicting types for item {key} {dkey} vs. {dname}. Data will be lost.")

            data_dict = alldicts[dname]

            # use a get() to account for first struct
            data_list = data_dict.get(key, [])
            data_list.append(item)
            data_dict[key] = data_list

    if len(new_items) > 0:
        warnings.warn(f"These items were not found in every struct: {new_items}")

    final_dict = {}
    for name, fa_list in alldicts["Array"].items():
        # make sure all dimensions are the same, possibly perform an hstack
        dims = list({f.ndim for f in fa_list})
        if len(dims) == 1:
            dims = dims[0]
            if dims == 1:
                stack_func = hstack
            elif dims == 2:
                stack_func = np.vstack
            else:
                raise ValueError(f"Can only stack arrays of one or two dimensions. Got {dims} dimensions.")
        else:
            raise ValueError(f"Found conflicting dimensions {dims} for item {name} in multiple structs.")
        try:
            final_dict[name] = stack_func(fa_list)
        except Exception:
            raise TypeError(
                f"Could not perform {stack_func} on arrays for column {name}. Arrays had shapes {[f.shape for f in fa_list]}"
            )

    # hstack lists of the same type - rt.hstack will route these to the correct routine
    for typename, typedict in alldicts.items():
        for name, itemlist in typedict.items():
            final_dict[name] = hstack(itemlist)

    result = TypeRegister.Struct(final_dict)
    # columns in final struct will appear in order of first occurrence in struct list
    order = list(all_keys.keys())
    result.col_move_to_front(order)
    return result


# --------------------------------------------------------------------------
def _hstack_dataset(ds_list: Union[list, tuple], destroy: bool = False):
    """
    Stacks columns from multiple datasets.

    If a dataset is missing a column that appears in others, it will fill the gap with the invalid for that column's dtype.
    Categoricals will be merged and stacked.
    Column types will be checked to make sure they can be safely stacked - no general type mismatch allowed.
    Columns of the same name must have the same number of dimension in each dataset (1 or 2 dimensions allowed)

    Parameters
    ----------
    ds_list : list of Dataset or tuple of Dataset

    Other Parameters
    ----------------
    destroy: bool, False
        Set to True to destroy the contents of the dataset to save on memory.
        !! This is dangerous so make sure you do not want the data anymore in the original datasets.
    """

    if not isinstance(ds_list, (list, tuple)):
        raise TypeError(f"Input must be a list or tuple of datasets. Got {type(ds_list)}")

    num_datasets = len(ds_list)
    if num_datasets == 0:
        # empty dict
        return TypeRegister.Dataset({})
    if num_datasets == 1:
        return ds_list[0]

    # collect dtypes, dimensions, shapes from all columns with given name
    types = {}
    ndims = {}
    shapes = {}
    for ds in ds_list:
        for name, col in ds.items():
            typelist = types.setdefault(name, [])
            # array subclasses should use their type, not their dtype
            if TypeRegister.is_array_subclass(col):
                typelist.append(type(col))
            else:
                typelist.append(col.dtype)

            # collecting number of dimensions for now
            # certain concatenations of different dimensions are allowed in numpy
            # we will raise an error
            dimlist = ndims.setdefault(name, [])
            dimlist.append(col.ndim)

            # for generating correct invalid before stack (only necessary for 2-dims)
            if col.ndim == 2:
                shapelist = shapes.setdefault(name, [])
                shapelist.append(col.shape[1])

    # check to make sure every column can be safely h-stacked
    for name, dts in types.items():
        types[name] = set(dts)
        types[name] = [dt for dt in types[name]]

        # matching types or compatible dtypes
        if len(types[name]) > 1:
            try:
                # possibly all numeric
                safe = all([t.char in NumpyCharTypes.AllInteger + NumpyCharTypes.AllFloat for t in types[name]])
                if not safe:
                    # possibly all string
                    safe = all([t.char in "US" for t in types[name]])
                # if conflicting warn the user
                if not safe:
                    warnings.warn(f"Mixing numeric with string types in column {name}. Unexpected results may occur.")
            # possibly a binned class
            except Exception:
                raise TypeError(
                    f"Custom class type was mixed with regular dtype in {name} column.  Types are {types[name]}."
                )

        # matching number of dimensions
        ndims[name] = list(set(ndims[name]))
        if len(ndims[name]) > 1:
            raise ValueError(f"Found multiple dimensions for column {name}: {ndims[name]}")
        else:
            dim = ndims[name][0]
            if dim != 1 and dim != 2:
                raise NotImplementedError(f"concat rows only supports columns of 1 or 2 dimensions, got {ndims[name]}")
            else:
                # two dimensional shape[1] matches (catch before numpy fails/invalids are generated later)
                if dim == 2:
                    width = list(set(shapes[name]))
                    if len(width) > 1:
                        raise ValueError(f"Found mismatch in axis-1 length for {name}, got {width}")
                    shapes[name] = width[0]
                ndims[name] = dim

    # we have our own hstack to avoid multiple copies when using safe types
    newcols = {}

    if destroy:
        # temporarily stop recycling (1=off)
        recycle_mode = TypeRegister.FastArray._ROFF(quiet=True)
        # force memory cleanup now
        rc.RecycleGarbageCollectNow(0)

    for name, t in types.items():
        # for each column, attempt to get it from each dataset, or create an invalid one
        column_list = []
        coldim = ndims[name]
        if coldim == 1:
            inv_count = 0
        else:
            inv_count = (0, shapes[name])
        # is_binned = t[0] in binned_types
        for ds in ds_list:
            nrows = ds._nrows
            # column was found in current dataset
            if name in ds:
                col = ds[name]

                # might add gap column, reset the running invalid column
                inv_count, column_list = _possibly_add_concat_gap(coldim, inv_count, col, column_list, t)

                # add the existing column
                column_list.append(col)

            # reserve room for the gap column, just add to gap length to minimize invalid array generation
            else:
                if coldim == 1:
                    inv_count += nrows
                else:
                    inv_count = (inv_count[0] + nrows, inv_count[1])

        # check to make sure there's not a leftover invalid column to be added
        # note: col is the last col value from the loop above
        inv_count, column_list = _possibly_add_concat_gap(coldim, inv_count, col, column_list, t)
        del col

        # 2-dimensions need to be handled differently
        if coldim == 1:
            stackfunc = hstack
        else:
            # NOTE: if riptable takes over vstack, remove the dtype fixup that happens in _possibly_add_concat_gap
            stackfunc = np.vstack

        # add to final dataset dict
        newcols[name] = stackfunc(column_list)
        del column_list

        # check if destroy the data inside the datasets
        if destroy:
            for ds in ds_list:
                # column was found in current dataset
                if name in ds:
                    del ds[name]

    returnds = TypeRegister.Dataset(newcols)
    del newcols

    if destroy:
        # turn recycling mode back to what it was
        if recycle_mode:
            TypeRegister.FastArray._RON(quiet=True)

    return returnds


# --------------------------------------------------------------------------
def _possibly_add_concat_gap(coldim: int, inv_count, col, column_list, types):
    """
    Called by _hstack_dataset
    Column existed in calling loop's current dataset, might need to create/add an invalid column
    before adding an existing column. If invalid was added, reset the invalid count.

    Parameters
    ----------
    coldim      : dimensions of column (1 or 2)
    inv_count   : running count of invalid gap. will be single int value for 1dim length or tuple of 2dim shape
    col         : existing column, used to generate the correct invalid value for the gap
    column_list : running list of all columns existing/invalid for final stack
    types       : dtypes of arrays - 2dim may need to generate invalid differently, 1dim will be fixed in hstack

    """
    if coldim == 1:
        if inv_count > 0:
            column_list.append(col.fill_invalid(shape=inv_count, inplace=False))
            inv_count = 0
    else:
        if inv_count[0] > 0:
            dtype = None
            if len(types) > 1:
                common = np.find_common_type(types, [])
                # numeric, supported, ensure final vstack result dtype - strings will be automatic
                if common.num <= 13:
                    # future optimization: store this for performance, reduce types list to final type
                    dtype = common
            column_list.append(col.fill_invalid(shape=inv_count, inplace=False, dtype=dtype))
            inv_count = (0, inv_count[1])

    return inv_count, column_list


# ------------------------------------------------------------
def _hstack_categorical(cats: list, verbose: bool = False, destroy: bool = False):
    """
    HStack Categoricals.

    The unique categories will be merged into a new unique list.
    The indices will be fixed to point to the new category array.
    The indices are hstacked and a new categorical is returned.

    Parameters
    ----------
    cats : list of Categorical
        Cats must be a list of categoricals.
    verbose : bool
        Enable verbose output. Defaults to False.
    destroy : bool
        This parameter is ignored by this function.

    Returns
    -------
    Categorical

    Examples
    --------
    >>> c1 = Categorical(['a','b','c'])
    >>> c2 = Categorical(['d','e','f'])
    >>> combined = Categorical.hstack([c1,c2])
    >>> combined
    Categorical([a, b, c, d, e, f]) Length: 6
        FastArray([1, 2, 3, 4, 5, 6]) Base Index: 1
        FastArray([b'a', b'b', b'c', b'd', b'e', b'f'], dtype='|S1') Unique count: 6
    """

    def attrs_match(attrlist, name):
        # ensure certain attributes are the same for all categoricals being stacked
        attrs = set(attrlist)
        if len(attrs) != 1:
            raise TypeError(
                f"hstack found {len(attrlist)} different values of the '{name}' attribute in provided Categoricals. Must all be the same."
            )
        return list(attrs)[0]

    # collect all the categorical modes and all the base indexes
    modes = []
    bases = []
    for cat in cats:
        if not isinstance(cat, TypeRegister.Categorical):
            raise TypeError(f"Categorical hstack is for categoricals, not {type(cat)}")
        # if cat.base_index not in (1, None):
        #    raise TypeError(f"only categoricals with base index 1 can be merged (to preserve invalid values).")
        modes.append(cat.category_mode)
        bases.append(cat.base_index)

    # all categoricals must be in same mode and have same base index
    mode = attrs_match(modes, "mode")
    base_index = attrs_match(bases, "base index")

    # the first categorical determines the ordered kwarg
    ordered = cats[0].ordered
    sort_display = cats[0].sort_gb

    # ==========================
    # todo: see _multistack_categoricals int rt_sds.py
    # stack indices
    # this will stack the fastarrays
    indices = rc.HStack(cats)
    idx_cutoffs = TypeRegister.FastArray([len(c._fa) for c in cats], dtype=np.int64).cumsum()

    # ------------------------- start rebuild here
    if mode in (CategoryMode.Dictionary, CategoryMode.IntEnum):

        # -----------------------
        # use info from grouping objects to stack
        glist = [c.grouping for c in cats]

        underlying = hstack([[*g._grouping_dict.values()][0] for g in glist])
        # stack all unique string arrays
        listnames = hstack([g._enum.category_array for g in glist])

        # collect, measure, stack integer arrays
        listcodes = [g._enum.code_array for g in glist]
        unique_cutoffs = [TypeRegister.FastArray([len(c) for c in listcodes], dtype=np.int64).cumsum()]
        listcodes = hstack(listcodes)

        # send in as two arrays
        listcats = [listcodes, listnames]

        # -----------------------
        base_index = None
        indices, listcats = merge_cats(
            indices, listcats, unique_cutoffs=unique_cutoffs, from_mapping=True, ordered=ordered, verbose=verbose
        )
        # TJD added check
        code = listcats[0][0]
        if isinstance(code, (int, np.integer)):
            # EXCEPT first value is string, and second is int
            newcats = dict(zip(listcats[1], listcats[0]))
        else:
            newcats = dict(zip(listcats[0], listcats[1]))

    else:
        category_dict = {}
        # now we need stack the unique cats
        for c in cats:

            # it might be multikey
            for i, v in enumerate(c.category_dict.values()):
                cv = category_dict.get(i, None)
                if cv is None:
                    category_dict[i] = [v]
                else:
                    cv.append(v)
                    category_dict[i] = cv

        listcats = []
        lastv = []
        for v in category_dict.values():
            listcats.append(hstack(v))
            lastv = v

        unique_cutoffs = [TypeRegister.FastArray([len(v) for v in lastv], dtype=np.int64).cumsum()]

        indices, newcats = merge_cats(
            indices,
            listcats,
            idx_cutoffs=idx_cutoffs,
            unique_cutoffs=unique_cutoffs,
            verbose=verbose,
            base_index=base_index,
            ordered=ordered,
        )

    newcats = TypeRegister.Grouping(
        indices, categories=newcats, _trusted=True, base_index=base_index, ordered=ordered, sort_display=sort_display
    )
    result = TypeRegister.Categorical(newcats)
    return result


# ------------------------------------------------------------
def _hstack_datetimenano(dtlist, destroy=False):
    """
    Performs an hstack on a list of DateTimeNano objects.
    All items in list must have their display set to the same timezone.
    NOTE: destroy ignored
    """
    # make sure all of the date time nano objects are set to be displayed relative to the same timezone
    timezone = dtlist[0]._timezone._timezone_str
    for dt in dtlist:
        if not isinstance(dt, TypeRegister.DateTimeNano):
            raise TypeError(f"Items to be hstacked must be DateTimeNano objects.")
        if dt._timezone._timezone_str != timezone:
            raise NotImplementedError(f"Can only hstack DateTimeNano objects in the same timezone.")
    if len(dtlist) == 1:
        return dtlist

    # hstack int64 utc nano arrays
    arr = rc.HStack(dtlist)

    # reconstruct with first item
    return TypeRegister.DateTimeNano.newclassfrominstance(arr, dtlist[0])


# ------------------------------------------------------------
def _hstack_timespan(tspans, destroy=False):
    """
    TODO: maybe add type checking?
    This is a very simple class, rewrap the hstack result in class.
    NOTE: destroy ignored
    """
    ts = rc.HStack(tspans)
    return TypeRegister.TimeSpan(ts)


# ------------------------------------------------------------
def _hstack_date(dates, destroy=False):
    """
    NOTE: destroy ignored
    """
    return _hstack_date_internal(dates, subclass=TypeRegister.Date)


# ------------------------------------------------------------
def _hstack_datespan(dates, destroy=False):
    """
    NOTE: destroy ignored
    """
    return _hstack_date_internal(dates, subclass=TypeRegister.DateSpan)


# ------------------------------------------------------------
def _hstack_date_internal(dates, subclass=TypeRegister.Date):
    """
    hstacks Date / DateSpan objects and returns a new Date / DateSpan object.

    Will be called by riptable.hstack() if the first item in the sequence is a Date object.

    Parameters
    ----------
    dates : list or tuple of Date / DateSpan objects

    Examples
    --------
    >>> d1 = Date('2015-02-01')
    >>> d2 = Date(['2016-02-01', '2017-02-01', '2018-02-01'])
    >>> hstack([d1, d2])
    Date([2015-02-01, 2016-02-01, 2017-02-01, 2018-02-01])
    """
    if len(dates) == 0:
        return subclass([])
    for d in dates:
        if not isinstance(d, subclass):
            # maybe extend this to support stacking with regular DateTimeNano objects?
            raise TypeError(f"Could not perform Date.hstack() on item of type {type(d)}")
    if len(dates) == 1:
        return dates

    stacked = rc.HStack(dates)
    return subclass.newclassfrominstance(stacked, dates[0])


# create a stack_rows alias for now
stack_rows = hstack_any



__all__ = ["GroupbyNumba"]

import functools
import numba as nb
import numpy as np
from typing import Any
from .config import get_global_settings
from .rt_enum import (
    GB_FUNC_NUMBA,
    GB_FUNCTIONS,
    GB_PACKUNPACK,
    INVALID_DICT,
    TypeRegister,
)
from .rt_groupbyops import GroupByOps
from .rt_numpy import empty, empty_like
from .numba.invalid_values import get_invalid, is_valid
from .rt_fastarraynumba import _check_fill_values

# NOTE YOU MUST INSTALL tbb
# conda install tbb
# to confirm...
# >>> from numba import threading_layer()
# >>> threading_layer()
# >>> 'tbb'
#
# See Table at end
# -------------------------------------------------------------------------------------------------
@nb.njit(cache=get_global_settings().enable_numba_cache, nogil=True)
def build_core_list(cores, unique_rows, binLow, binHigh):
    dividend = unique_rows // cores
    remainder = unique_rows % cores
    high = 0
    low = 0

    for i in range(cores):
        # Calculate band range
        high = low + dividend

        # add in any remainder until nothing left
        if remainder > 0:
            high += 1
            remainder -= 1

        binLow[i] = low
        binHigh[i] = high

        # next low bin is the previous high bin
        low = high


class GroupbyNumba(GroupByOps):

    # how many cores to cap the computation at
    # NOTE: this is not how many cores the system has but a number where
    # we believe thrashing takes place.  This number could be dynamic per algo in the future.
    CORE_COUNT = 12

    # -------------------------------------------------------------------------------------------------
    def _nb_groupbycalculateall(values, ikey, unique_rows, funcList, binLowList, binHighList, func_param):
        results = []

        unique_rows += 1
        corecount = min(GroupbyNumba.CORE_COUNT, unique_rows)

        binLow = np.empty(corecount, dtype=np.int32)
        binHigh = np.empty(corecount, dtype=np.int32)

        build_core_list(corecount, unique_rows, binLow, binHigh)

        for funcnum, inputdata in zip(funcList, values):
            nbrev = NUMBA_REVERSE_TABLE[funcnum]

            # lookup function to call
            nbfunc = nbrev["func_back"]

            # lookup return dtype requested
            dtypefunc = nbrev["func_dtype"]
            if dtypefunc is None:
                # output is same dtype as input
                dtype = inputdata.dtype
            else:
                dtype = dtypefunc(inputdata.dtype)

            # allocate for numba
            ret = empty(unique_rows, dtype=dtype)

            nbfunc(ikey, unique_rows, binLow, binHigh, inputdata, ret, *func_param)
            results.append(ret)
        return results

    # -------------------------------------------------------------------------------------------------
    # This routine is called before the numba routines
    def _nb_groupbycalculateallpack(
        values,  # list of arrays (the data to be calculated)
        ikey,  # bin numbers (integer array)
        iGroup,  # used to go over
        iFirstGroup,
        nCountGroup,
        unique_rows,  # often the same as len(iFirstGroup)
        funcList,  # support aggregation
        binLowList,  # start bin to work on for prange
        binHighList,  # high bin to work on for prange
        inplace,  # if to apply function inplace
        func_param,
    ):  # parameters

        results = []

        # TODO: add enumerate here
        for funcnum, inputdata in zip(funcList, values):
            nbrev = NUMBA_REVERSE_TABLE[funcnum]

            # lookup function to call
            nbfunc = nbrev["func_back"]

            # lookup return dtype requested
            dtypefunc = nbrev["func_dtype"]
            if dtypefunc is None:
                # output is same dtype as input
                dtype = inputdata.dtype
            else:
                dtype = dtypefunc(inputdata.dtype)

            ret = None
            if not inplace:
                # allocate for numba
                ret = empty(len(inputdata), dtype=dtype)

            nbfunc(iGroup, iFirstGroup, nCountGroup, binLowList[0], binHighList[0], inputdata, ret, *func_param)

            if inplace:
                results.append(inputdata)
            else:
                results.append(ret)

        return results

    # -------------------------------------------------------------------------------------------------
    @nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)
    def _numbasum(ikey, unique_rows, binLow, binHigh, data, ret):
        datacount = len(ikey)

        # binLow and binHigh are arrays (same length)
        # they divide up the work for prange while also allowing selective group filtering
        for core in nb.prange(len(binLow)):
            binlow = binLow[core]
            binhigh = binHigh[core]

            # zero out summation counters before we begin
            for i in range(binlow, binhigh):
                ret[i] = 0

            # concurrently loop over all the data
            for index in range(datacount):
                grpIdx = ikey[index]

                # make sure assigned to our range (concurrency issue)
                if grpIdx >= binlow and grpIdx < binhigh:
                    ret[grpIdx] += data[index]

    # -------------------------------------------------------------------------------------------------
    @nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)
    def _numbamin(ikey, unique_rows, binLow, binHigh, data, ret):
        inv = INVALID_DICT[ret.dtype.num]
        datacount = len(ikey)

        for core in nb.prange(len(binLow)):
            binlow = binLow[core]
            binhigh = binHigh[core]

            # mark all initial values as invalid we begin
            for i in range(binlow, binhigh):
                ret[i] = inv

            # concurrently loop over all the data
            for index in range(datacount):
                grpIdx = ikey[index]

                # make sure assigned to our range (concurrency issue)
                if grpIdx >= binlow and grpIdx < binhigh:
                    val = data[index]

                    # set the min, use not >= to handle nan comparison
                    if ret[grpIdx] == inv or not val >= ret[grpIdx]:
                        ret[grpIdx] = val

    # -------------------------------------------------------------------------------------------------
    @nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)
    def _numbaEMA(iGroup, iFirstGroup, nCountGroup, binLow, binHigh, data, ret, time, decayRate):
        for grpIdx in nb.prange(binLow, binHigh):
            start = iFirstGroup[grpIdx]
            last = start + nCountGroup[grpIdx]

            # init per group data
            lastEma = 0.0
            lastTime = time[iGroup[start]]
            for index in range(start, last):
                rowIdx = iGroup[index]

                # ema calculation
                timeDelta = time[rowIdx] - lastTime
                lastTime = time[rowIdx]
                lastEma = data[rowIdx] + lastEma * np.exp(-decayRate * timeDelta)

                # store the return result
                ret[rowIdx] = lastEma

    # -------------------------------------------------------------------------------------------------#
    @nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)
    def _numbaEMA2(iGroup, iFirstGroup, nCountGroup, data, ret, time, decayRate):
        """
        For each group defined by the grouping arguments, sets 'ret' to a true EMA of the 'data'
        argument using the time argument as the time and the 'decayRate' as the decay rate.

        Arguments
        ---------
        iGroup, iFirstGroup, nCountGroup:  from a groupby object's 'get_groupings' method
        data:  the original data to be opperated on
        ret:  a blank array the same size as 'data' which will return the processed data
        time: a list of times associated to the rows of data
        decayRate: the decay rate (e based)

        TODO:  Error checking.
        """
        for grpIdx in nb.prange(1, iFirstGroup.shape[0]):
            startIdx = iFirstGroup[grpIdx]
            nInGrp = nCountGroup[grpIdx]
            endIdx = startIdx + nInGrp
            rowIdx = iGroup[startIdx:endIdx]

            if nInGrp > 0:
                rows = data[rowIdx]
                times = time[rowIdx]
                totalWeight = 0.0
                totalValues = 0.0
                pEMA = np.nan
                pTime = times[0]
                for (idx, (t, v)) in enumerate(zip(times, rows)):
                    if not np.isnan(v):
                        deltaT = t - pTime
                        decay = np.exp(-decayRate * deltaT)
                        totalWeight = totalWeight * decay + 1
                        totalValues = totalValues * decay + v
                        pTime = t
                        pEMA = totalValues / totalWeight
                    rows[idx] = pEMA

                ret[rowIdx] = rows
        return

    ### Trim (an example which returns a dataset the same size as the original) ###
    # -------------------------------------------------------------------------------------------------#
    @nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)
    def _numbaTrim(iGroup, iFirstGroup, nCountGroup, data, ret, x, y):
        """
        For each group defined by the grouping arguments, sets 'ret' to be a copy of the 'data'
        with elements below the 'x'th percentile or above the 'y'th percentile of the group set to nan.

        Arguments
        ---------
        iGroup, iFirstGroup, nCountGroup:  from a groupby object's 'get_groupings' method
        data:  the original data to be opperated on
        ret:  a blank array the same size as 'data' which will return the processed data
        x:  the lower percentile bound
        y:  the upper percentile bound
        """
        for grpIdx in nb.prange(1, iFirstGroup.shape[0]):
            startIdx = iFirstGroup[grpIdx]
            endIdx = startIdx + nCountGroup[grpIdx]
            rowIdx = iGroup[startIdx:endIdx]
            rows = data[rowIdx]
            (a, b) = np.nanpercentile(rows, [x, y])
            mask = (rows <= a) | (rows >= b)
            rows[mask] = np.nan
            ret[rowIdx] = rows
        return

    def grpTrim(grp, x, y):
        """
        For each column, for each group, determine the x'th and y'th percentile of the data
        and set data below the x'th percentile or above the y'th percentile to nan.

        Arguments
        ---------
        grp:  a groupby object
        x:  lower percentile
        y:  uppper percentile

        Return
        ------
        A dataset with the values outside the given percentiles set to np.nan

        TODO:  Test column types to make sure that the numba code will work nicely
        """
        g = grp.get_groupings()
        iGroup = g["iGroup"]
        iFirstGroup = g["iFirstGroup"]
        nCountGroup = g["nCountGroup"]

        # retData = rt.Dataset(tmp.grp.gbkeys)
        retData = grp._dataset[list(grp.gbkeys.keys())]
        for colName in grp._dataset:
            if colName not in grp.gbkeys.keys():
                ret = empty(grp._dataset.shape[0])
                grp._numbaTrim(iGroup, iFirstGroup, nCountGroup, grp._dataset[colName], ret, x, y)
                retData[colName] = ret
        return retData

    # FillForward
    # -------------------------------------------------------------------------------------------------#

    def nb_fill_forward(self, *args, limit=0, fill_val=None, inplace=False):
        """
        Replace NaN and invalid array values by propagating the last encountered valid
        group value forward.

        Optionally, you can modify the original array if it's not locked.

        Parameters
        ----------
        *args : array or list of arrays
            The array or arrays that contain NaN or invalid values you want to replace.
        limit : int, default 0 (disabled)
            The maximium number of consecutive NaN or invalid values to fill. If there
            is a gap with more than this number of consecutive NaN or invalid values,
            the gap will be only partially filled. If no `limit` is specified, all
            consecutive NaN and invalid values are replaced.
        fill_val : scalar, default None
            The value to use where there is no valid group value to propagate forward.
            If `fill_val` is not specified, NaN and invalid values aren't replaced where
            there is no valid group value to propagate forward.
        inplace: bool, default False
            If False, return a copy of the array. If True, modify original data. This
            will modify any other views on this object. This fails if the array is
            locked.

        Returns
        -------
        Dataset-like object
            The dataset (categorical) will be the same size and have the same dtypes as the
            original input.
        """
        # Lookup our function to get a function_number
        return self._calculate_all(
            NUMBA_REVERSE_FUNC[GroupbyNumba.nb_fill_forward], *args, func_param=(fill_val, limit), inplace=inplace
        )

    def nb_fill_backward(self, *args, fill_val, limit=0, inplace=False):
        """
        Replace NaN and invalid array values by propagating the next encountered valid
        group value backward.

        Optionally, you can modify the original array if it's not locked.

        Parameters
        ----------
        *args : array or list of arrays
            The array or arrays that contain NaN or invalid values you want to replace.
        limit : int, default 0 (disabled)
            The maximium number of consecutive NaN or invalid values to fill. If there
            is a gap with more than this number of consecutive NaN or invalid values,
            the gap will be only partially filled. If no `limit` is specified, all
            consecutive NaN and invalid values are replaced.
        fill_val : scalar, default None
            The value to use where there is no valid group value to propagate backward.
            If `fill_val` is not specified, NaN and invalid values aren't replaced where
            there is no valid group value to propagate backward.
        inplace: bool, default False
            If False, return a copy of the array. If True, modify original data. This
            will modify any other views on this object. This fails if the array is
            locked.

        Returns
        -------
        Dataset-like object
            The dataset (categorical) will be the same size and have the same dtypes as the
            original input.
        """
        # Lookup our function to get a function_number
        return self._calculate_all(
            NUMBA_REVERSE_FUNC[GroupbyNumba.nb_fill_backward], *args, func_param=(fill_val, limit), inplace=inplace
        )

    @staticmethod
    def _nb_fill_backend(iGroup, iFirstGroup, nCountGroup, binLow, binHigh, data, ret, fill_val, limit, direction):
        """
        Numba backend implementation for grouped fill_forward and fill_backward for all aplicable dtypes.

        Arguments
        ---------
        iGroup, iFirstGroup, nCountGroup : np.ndarray
            Arrays from a groupby object's 'get_groupings' method
        binLow, binHigh : int
            Indexes corresponding to the first and the last groups in iFirstGroup and nCountGroup
        data :  array
            The original data to be opperated on
        ret :  array
            An empty array the same size as 'data' which will contain the processed data.
            Must be `None` for inplace operation
        fill_val, limit : parameters for nb_fill_forward/nb_fill_backward
            The value to use where there is no valid group value to propagate forward/backward.
            If `fill_val` is not specified, NaN and invalid values aren't replaced where
            there is no valid group value to propagate forward/backward.
        direction:  int (-1 or 1)
            direction = 1 corresponds to fill_forward, -1 corresponds to fill_backward
        """

        dtype = data.dtype
        if not np.issubdtype(dtype, np.floating) and not np.issubdtype(dtype, np.integer):
            raise TypeError(f"Filling for type `{dtype}` is currently not supported.")

        if fill_val is None:
            fill_val = get_invalid(data)

        fill_val = dtype.type(fill_val)

        if ret is None:
            # inplace, just pass data as ret itself
            ret = data

        if limit is None:
            limit = 0
        elif limit < 0:
            raise TypeError(f"The limit kwarg cannot be less than 0.")

        direction = np.int8(direction)
        limit = np.int64(limit)
        binLow = np.int64(binLow)
        binHigh = np.int64(binHigh)

        GroupbyNumba._numba_fill_direction(
            direction,
            iGroup,
            iFirstGroup,
            nCountGroup,
            binLow,
            binHigh,
            data,
            ret,
            fill_val,
            limit,
        )

        return

    @staticmethod
    @nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)
    def _numba_fill_direction(
        direction: np.int8,
        iGroup: np.ndarray,
        iFirstGroup: np.ndarray,
        nCountGroup: np.ndarray,
        binLow: np.int64,
        binHigh: np.int64,
        data: np.ndarray,
        ret: np.ndarray,
        fill_val: Any,
        limit: np.int64,
    ):
        for grpIdx in nb.prange(binLow, binHigh):
            start = iFirstGroup[grpIdx]
            last = start + nCountGroup[grpIdx]

            if direction == 1:
                idx_range = range(start, last)
            else:
                idx_range = range(last - 1, start - 1, -1)

            last_valid = fill_val
            if limit <= 0:
                for index in idx_range:
                    rowIdx = iGroup[index]
                    if is_valid(data[rowIdx]):
                        last_valid = data[rowIdx]
                    ret[rowIdx] = last_valid

            else:
                counter = limit
                for index in idx_range:
                    rowIdx = iGroup[index]
                    if not is_valid(data[rowIdx]):
                        if counter > 0:
                            ret[rowIdx] = last_valid
                            counter -= 1
                        else:
                            # copy invalid value
                            ret[rowIdx] = data[rowIdx]
                    else:
                        counter = limit
                        last_valid = data[rowIdx]
                        ret[rowIdx] = last_valid
        return

    @nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)
    def _numbaFillForward(iGroup, iFirstGroup, nCountGroup, data, ret):
        """
        propogate forward non-NaN values within a group, overwriting NaN values.
        TODO:  better documentation
        """
        for grpIdx in nb.prange(1, iFirstGroup.shape[0]):
            startIdx = iFirstGroup[grpIdx]
            endIdx = startIdx + nCountGroup[grpIdx]
            rowIdx = iGroup[startIdx:endIdx]
            rows = data[rowIdx]

            fill = np.nan
            for idx in range(rows.shape[0]):
                if np.isnan(rows[idx]):
                    rows[idx] = fill
                else:
                    fill = rows[idx]
            ret[rowIdx] = rows
        return

    @nb.njit(parallel=True, cache=get_global_settings().enable_numba_cache, nogil=True)
    def _numbaFillBackward(iGroup, iFirstGroup, nCountGroup, data, ret):
        """
        propogate backward non-NaN values within a group, overwriting NaN values.
        TODO:  better documentation
        """
        for grpIdx in nb.prange(1, iFirstGroup.shape[0]):
            startIdx = iFirstGroup[grpIdx]
            endIdx = startIdx + nCountGroup[grpIdx]
            rowIdx = iGroup[startIdx:endIdx]
            rows = data[rowIdx]

            fill = np.nan
            for idx in range(rows.shape[0]):
                if np.isnan(rows[-idx - 1]):
                    rows[-idx - 1] = fill
                else:
                    fill = rows[-idx - 1]
            ret[rowIdx] = rows
        return

    def grpFillForward(grp):
        """
        propogate forward non-NaN values within a group, overwriting NaN values.
        TODO:  better documentation
        """
        g = grp.get_groupings()
        iGroup = g["iGroup"]
        iFirstGroup = g["iFirstGroup"]
        nCountGroup = g["nCountGroup"]

        # retData = rt.Dataset(tmp.grp.gbkeys)
        retData = grp._dataset[list(grp.gbkeys.keys())]
        for colName in grp._dataset:
            if colName not in grp.gbkeys.keys():
                ret = empty(grp._dataset.shape[0])
                grp._numbaFillForward(iGroup, iFirstGroup, nCountGroup, grp._dataset[colName], ret)
                retData[colName] = ret
        return retData

    def grpFillBackward(grp):
        """
        propogate backward non-NaN values within a group, overwriting NaN values.
        TODO:  better documentation
        """
        g = grp.get_groupings()
        iGroup = g["iGroup"]
        iFirstGroup = g["iFirstGroup"]
        nCountGroup = g["nCountGroup"]

        # retData = rt.Dataset(tmp.grp.gbkeys)
        retData = grp._dataset[list(grp.gbkeys.keys())]
        for colName in grp._dataset:
            if colName not in grp.gbkeys.keys():
                ret = empty(grp._dataset.shape[0])
                grp._numbaFillBackward(iGroup, iFirstGroup, nCountGroup, grp._dataset[colName], ret)
                retData[colName] = ret
        return retData

    def grpFillForwardBackward(grp):
        """
        propogate forward, then backward, non-NaN values within a group, overwriting NaN values.
        TODO:  better documentation
        """
        g = grp.get_groupings()
        iGroup = g["iGroup"]
        iFirstGroup = g["iFirstGroup"]
        nCountGroup = g["nCountGroup"]

        # retData = rt.Dataset(tmp.grp.gbkeys)
        retData = grp._dataset[list(grp.gbkeys.keys())]
        for colName in grp._dataset:
            if colName not in grp.gbkeys.keys():
                forwardFilled = empty(grp._dataset.shape[0])
                grp._numbaFillForward(iGroup, iFirstGroup, nCountGroup, grp._dataset[colName], forwardFilled)
                ret = empty(grp._dataset.shape[0])
                grp._numbaFillBackward(iGroup, iFirstGroup, nCountGroup, forwardFilled, ret)
                retData[colName] = ret
        return retData

    # ---------------------------------------------------------------
    def nb_ema(self, *args, time=None, decay_rate=None, **kwargs):
        """
        Other Parameters
        ----------------
        time: an array of times (often in nanoseconds) associated to the rows of data
        decayRate: the scalar decay rate (e based)
        """
        if time is None:
            raise KeyError("time cannot be none")
        if len(time) != self._dataset.shape[0]:
            raise TypeError(f"time array must be the same size as the dataset")

        if decay_rate is None:
            raise KeyError("decay_rate cannot be none")
        if not np.isscalar(decay_rate):
            raise TypeError(f"decay_rate must be a scalar not type {type(decay_rate)}")

        # Lookup our function to get a function_number
        return self._calculate_all(
            NUMBA_REVERSE_FUNC[GroupbyNumba.nb_ema], *args, func_param=(time, decay_rate), **kwargs
        )

    # ---------------------------------------------------------------
    def nb_sum_punt_test(self, *args, **kwargs):
        """Compute sum of group"""
        return self._calculate_all(GB_FUNCTIONS.GB_SUM, *args, **kwargs)

    # ---------------------------------------------------------------
    def nb_sum(self, *args, **kwargs):
        """Compute sum of group"""
        return self._calculate_all(NUMBA_REVERSE_FUNC[GroupbyNumba.nb_sum], *args, **kwargs)

    # ---------------------------------------------------------------
    def nb_min(self, *args, **kwargs):
        """Compute sum of group"""
        return self._calculate_all(NUMBA_REVERSE_FUNC[GroupbyNumba.nb_min], *args, **kwargs)


# ----------------------------------------------------
# add more routines here to determine the output dtype from the input dtype
def NUMBA_DTYPE_FLOATS(dtype):
    if isinstance(dtype, np.float64):
        return np.float64
    return np.float32


def NUMBA_DTYPE_SUM(dtype):
    # upcast most ints to int64
    if isinstance(dtype, np.uint64):
        return np.uint64
    if dtype.num <= 10:
        return np.int64
    return dtype


CALC_PACK = GroupbyNumba._nb_groupbycalculateallpack
CALC_UNPACK = GroupbyNumba._nb_groupbycalculateall

# ------------------------------------------------------------
#     name:       (basic/packing,       func_frontend,                   func_backend,          gb_function,  dtype,              return_full True/False)
#     ------      -------------------   ------------------------         ------------------     -----------   ----------------    ------------------
NUMBA_GB_TABLE = {
    "nb_ema": (GB_PACKUNPACK.PACK, GroupbyNumba.nb_ema, GroupbyNumba._numbaEMA, CALC_PACK, NUMBA_DTYPE_FLOATS, True),
    "nb_fill_forward": (
        GB_PACKUNPACK.PACK,
        GroupbyNumba.nb_fill_forward,
        functools.partial(GroupbyNumba._nb_fill_backend, direction=1),
        CALC_PACK,
        None,
        True,
    ),
    "nb_fill_backward": (
        GB_PACKUNPACK.PACK,
        GroupbyNumba.nb_fill_backward,
        functools.partial(GroupbyNumba._nb_fill_backend, direction=-1),
        CALC_PACK,
        None,
        True,
    ),
    "nb_sum": (GB_PACKUNPACK.UNPACK, GroupbyNumba.nb_sum, GroupbyNumba._numbasum, CALC_UNPACK, NUMBA_DTYPE_SUM, False),
    "nb_min": (GB_PACKUNPACK.UNPACK, GroupbyNumba.nb_min, GroupbyNumba._numbamin, CALC_UNPACK, None, False),
    "nb_sum_punt": (GB_PACKUNPACK.UNPACK, GroupbyNumba.nb_sum_punt_test, None, None, None, False),
}

# key is a function number
# key : (funcname, requirespacking, frontend, backend, grouper)

NUMBA_REVERSE_TABLE = {}
NUMBA_REVERSE_FUNC = {}

# start assigning funcnum values at 1000
for i, (k, v) in enumerate(NUMBA_GB_TABLE.items()):
    NUMBA_REVERSE_TABLE[i + GB_FUNC_NUMBA] = {
        "name": k,
        "packing": v[0],
        "func_front": v[1],
        "func_back": v[2],
        "func_gb": v[3],
        "func_dtype": v[4],
        "return_full": v[5],
    }
    NUMBA_REVERSE_FUNC[v[1]] = i + GB_FUNC_NUMBA

# register our custom functions
GroupByOps.register_functions(NUMBA_REVERSE_TABLE)
TypeRegister.Grouping.register_functions(NUMBA_REVERSE_TABLE)
